"""Domain types for the memory plugin."""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from enum import IntEnum
from typing import TypedDict

MEMORY_METADATA_SEPARATOR = "\n---\n"
MEMORY_SOURCE = "plugin-memory"
PLUGIN_MEMORY_TABLE = "plugin_memory"

IMPORTANCE_LABELS: dict[int, str] = {1: "low", 2: "normal", 3: "high", 4: "critical"}


class MemoryImportance(IntEnum):
    LOW = 1
    NORMAL = 2
    HIGH = 3
    CRITICAL = 4


@dataclass
class ParsedMemory:
    content: str
    tags: list[str]
    importance: MemoryImportance


@dataclass
class RememberParameters:
    content: str | None = None
    tags: list[str] = field(default_factory=list)
    importance: MemoryImportance = MemoryImportance.NORMAL
    metadata: dict[str, str | int | bool] = field(default_factory=dict)


@dataclass
class RecallParameters:
    query: str | None = None
    tags: list[str] = field(default_factory=list)
    limit: int = 10
    min_importance: MemoryImportance = MemoryImportance.LOW


@dataclass
class ForgetParameters:
    memory_id: str | None = None
    content: str | None = None


class MemorySearchResult(TypedDict):
    id: str
    content: str
    tags: list[str]
    importance: int
    created_at: int


def encode_memory_text(
    content: str, tags: list[str], importance: MemoryImportance
) -> str:
    """Encode memory content with metadata into a storable text format."""
    metadata = json.dumps({"t": tags, "i": int(importance)})
    return f"{metadata}{MEMORY_METADATA_SEPARATOR}{content}"


def decode_memory_text(text: str) -> ParsedMemory:
    """Decode a stored memory text into its content and metadata."""
    sep_index = text.find(MEMORY_METADATA_SEPARATOR)
    if sep_index == -1:
        return ParsedMemory(content=text, tags=[], importance=MemoryImportance.NORMAL)

    metadata_str = text[:sep_index]
    content = text[sep_index + len(MEMORY_METADATA_SEPARATOR) :]

    try:
        metadata = json.loads(metadata_str)
        raw_tags = metadata.get("t", [])
        tags = [str(t) for t in raw_tags] if isinstance(raw_tags, list) else []
        importance_val = metadata.get("i", 2)
        importance = (
            MemoryImportance(importance_val)
            if isinstance(importance_val, int) and 1 <= importance_val <= 4
            else MemoryImportance.NORMAL
        )
        return ParsedMemory(content=content, tags=tags, importance=importance)
    except (json.JSONDecodeError, ValueError):
        return ParsedMemory(content=text, tags=[], importance=MemoryImportance.NORMAL)



"""Base types for provider definitions following the elizaOS plugin pattern."""

from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import Protocol


class RuntimeProtocol(Protocol):
    def get_service(self, name: str) -> object | None: ...

    def get_memory_manager(self) -> object | None: ...


@dataclass
class ProviderResult:
    text: str
    data: dict[str, str | int | float | bool | list | dict | None] | None = None


ProviderFunc = Callable[
    [RuntimeProtocol, object, object], Coroutine[object, object, ProviderResult]
]


@dataclass
class Provider:
    name: str
    description: str
    get: ProviderFunc



"""Base types for action definitions following the elizaOS plugin pattern."""

from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import Protocol, TypedDict


class MemoryContent(TypedDict, total=False):
    text: str
    source: str


class Memory(TypedDict, total=False):
    id: str
    agentId: str
    roomId: str
    userId: str
    content: MemoryContent
    createdAt: int


class State(TypedDict, total=False):
    pass


class ActionResult(TypedDict, total=False):
    text: str
    success: bool
    data: dict[str, str | int | float | bool | list | dict | None]


class GetMemoriesParams(TypedDict, total=False):
    roomId: str
    tableName: str
    count: int


class RuntimeProtocol(Protocol):
    """Runtime protocol aligned with current DB API (create_memory, get_memories, delete_memory)."""

    @property
    def agent_id(self) -> str: ...

    def get_setting(self, key: str) -> str | None: ...

    def get_service(self, name: str) -> object | None: ...

    async def create_memory(
        self, memory: Memory, table_name: str, unique: bool = False
    ) -> str: ...

    async def get_memories(self, params: GetMemoriesParams) -> list[Memory]: ...

    async def delete_memory(self, memory_id: str) -> None: ...

    async def use_model(self, model_type: str, params: dict[str, str]) -> str | None: ...


HandlerCallback = Callable[[dict[str, str | list[str] | None]], Coroutine[object, object, None]]

ValidateFunc = Callable[[RuntimeProtocol, Memory, State | None], Coroutine[object, object, bool]]

HandlerFunc = Callable[
    [
        RuntimeProtocol,
        Memory,
        State | None,
        dict[str, str | list[str] | int | None] | None,
        HandlerCallback | None,
    ],
    Coroutine[object, object, ActionResult],
]


@dataclass
class ActionExample:
    name: str
    content: dict[str, str | list[str]]


@dataclass
class Action:
    name: str
    description: str
    similes: list[str]
    examples: list[list[ActionExample]]
    validate: ValidateFunc
    handler: HandlerFunc


def create_action(
    name: str,
    description: str,
    similes: list[str],
    examples: list[list[ActionExample]],
    validate: ValidateFunc,
    handler: HandlerFunc,
) -> Action:
    return Action(
        name=name,
        description=description,
        similes=similes,
        examples=examples,
        validate=validate,
        handler=handler,
    )



"""Plugin definition for the memory plugin."""

from dataclasses import dataclass, field

from elizaos_plugin_memory.actions import (
    forget_action,
    recall_action,
    remember_action,
)
from elizaos_plugin_memory.actions.base import Action
from elizaos_plugin_memory.providers import memory_context_provider
from elizaos_plugin_memory.providers.base import Provider


@dataclass
class Plugin:
    name: str
    description: str
    actions: list[Action]
    providers: list[Provider] = field(default_factory=list)


memory_plugin = Plugin(
    name="@elizaos/plugin-memory-py",
    description="Plugin for long-term memory management with remember, recall, and forget capabilities",
    actions=[
        remember_action,
        recall_action,
        forget_action,
    ],
    providers=[
        memory_context_provider,
    ],
)



"""FORGET action - remove a stored memory by ID or by matching content."""

from __future__ import annotations

import json
import logging
import re

from elizaos_plugin_memory.actions.base import (
    ActionExample,
    ActionResult,
    HandlerCallback,
    Memory,
    RuntimeProtocol,
    State,
    create_action,
)
from elizaos_plugin_memory.types import (
    MEMORY_SOURCE,
    PLUGIN_MEMORY_TABLE,
    decode_memory_text,
)

logger = logging.getLogger(__name__)


async def validate(
    runtime: RuntimeProtocol,
    _message: Memory,
    _state: State | None = None,
) -> bool:
    try:
        return callable(getattr(runtime, "get_memories", None)) and callable(
            getattr(runtime, "delete_memory", None)
        )
    except Exception:
        return False


async def handler(
    runtime: RuntimeProtocol,
    message: Memory,
    _state: State | None = None,
    options: dict[str, str | list[str] | int | None] | None = None,
    callback: HandlerCallback | None = None,
) -> ActionResult:
    try:
        content_data = message.get("content", {})
        content = content_data.get("text", "")
        if not content:
            error_message = "Please specify which memory to forget."
            if callback:
                await callback({"text": error_message, "source": content_data.get("source")})
            return {"text": error_message, "success": False}

        # Direct removal by ID
        memory_id = str(options.get("memoryId", "")) if options else ""
        if memory_id:
            await runtime.delete_memory(memory_id)
            success_message = f"Removed memory with ID: {memory_id}"
            if callback:
                await callback({"text": success_message, "source": content_data.get("source")})
            return {"text": success_message, "success": True, "data": {"removedId": memory_id}}

        # Search for matching memory
        memories = await runtime.get_memories(
            {
                "roomId": message.get("roomId", ""),
                "tableName": PLUGIN_MEMORY_TABLE,
                "count": 100,
            }
        )

        plugin_memories = [
            m for m in memories if m.get("content", {}).get("source") == MEMORY_SOURCE
        ]

        if not plugin_memories:
            no_mem_msg = "No stored memories found to remove."
            if callback:
                await callback({"text": no_mem_msg, "source": content_data.get("source")})
            return {"text": no_mem_msg, "success": True}

        # Use LLM to find the best matching memory
        search_content = str(options.get("content", content)) if options else content
        descriptions = []
        for i, m in enumerate(plugin_memories):
            parsed = decode_memory_text(m.get("content", {}).get("text", ""))
            descriptions.append(f'{i}: "{parsed.content}"')

        match_prompt = (
            "Given the user's request to forget a memory, identify which memory index matches.\n\n"
            f'User request: "{search_content}"\n\n'
            f"Available memories:\n{chr(10).join(descriptions)}\n\n"
            'Return ONLY a JSON object: {"index": <number or -1>, "confidence": <0.0 to 1.0>}'
        )

        response = await runtime.use_model("TEXT_LARGE", {"prompt": match_prompt})
        if not response:
            raise RuntimeError("Failed to identify memory to remove")

        cleaned = re.sub(r"^```(?:json)?\n?", "", response)
        cleaned = re.sub(r"\n?```$", "", cleaned).strip()
        match_result: dict[str, int | float] = json.loads(cleaned)

        idx = int(match_result.get("index", -1))
        confidence = float(match_result.get("confidence", 0.0))

        if idx < 0 or idx >= len(plugin_memories) or confidence < 0.5:
            no_match_msg = "Could not find a matching memory to remove. Please be more specific."
            if callback:
                await callback({"text": no_match_msg, "source": content_data.get("source")})
            return {"text": no_match_msg, "success": False}

        target = plugin_memories[idx]
        parsed = decode_memory_text(target.get("content", {}).get("text", ""))
        target_id = target.get("id", "")

        if target_id:
            await runtime.delete_memory(target_id)

        success_message = f'Removed memory: "{parsed.content}"'
        if callback:
            await callback({"text": success_message, "source": content_data.get("source")})

        return {
            "text": success_message,
            "success": True,
            "data": {"removedId": target_id, "content": parsed.content},
        }

    except Exception as error:
        logger.error("Failed to forget memory: %s", error)
        error_message = f"Failed to forget memory: {error}"
        if callback:
            await callback(
                {"text": error_message, "source": message.get("content", {}).get("source")}
            )
        return {"text": error_message, "success": False}


forget_action = create_action(
    name="FORGET",
    description="Remove a stored memory by ID or by matching content description",
    similes=["forget", "remove-memory", "delete-memory", "erase-memory", "clear-memory"],
    examples=[
        [
            ActionExample(
                name="User",
                content={"text": "Forget what you know about my favorite color."},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "I'll remove that memory about your favorite color.",
                    "actions": ["FORGET"],
                },
            ),
        ],
        [
            ActionExample(
                name="User",
                content={"text": "Delete the memory about the project deadline."},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "I've removed the memory about the project deadline.",
                    "actions": ["FORGET"],
                },
            ),
        ],
    ],
    validate=validate,
    handler=handler,
)



from elizaos_plugin_memory.plugin import memory_plugin
from elizaos_plugin_memory.types import (
    MEMORY_SOURCE,
    ForgetParameters,
    MemoryImportance,
    MemorySearchResult,
    ParsedMemory,
    RecallParameters,
    RememberParameters,
    decode_memory_text,
    encode_memory_text,
)

__all__ = [
    "memory_plugin",
    "MemoryImportance",
    "ParsedMemory",
    "RememberParameters",
    "RecallParameters",
    "ForgetParameters",
    "MemorySearchResult",
    "MEMORY_SOURCE",
    "encode_memory_text",
    "decode_memory_text",
]

__version__ = "2.0.0"



"""Tests for the memory plugin actions, types, and providers."""

from __future__ import annotations

from unittest.mock import AsyncMock

import pytest

from elizaos_plugin_memory.actions.forget import forget_action
from elizaos_plugin_memory.actions.recall import recall_action
from elizaos_plugin_memory.actions.remember import remember_action
from elizaos_plugin_memory.providers.memory_context import memory_context_provider
from elizaos_plugin_memory.types import (
    MEMORY_METADATA_SEPARATOR,
    MEMORY_SOURCE,
    MemoryImportance,
    decode_memory_text,
    encode_memory_text,
)


# --- Type Encoding / Decoding ---


class TestEncoding:
    def test_encode_decode_roundtrip(self) -> None:
        content = "My favorite color is blue"
        tags = ["preference", "color"]
        importance = MemoryImportance.HIGH

        encoded = encode_memory_text(content, tags, importance)
        decoded = decode_memory_text(encoded)

        assert decoded.content == content
        assert decoded.tags == tags
        assert decoded.importance == importance

    def test_decode_plain_text(self) -> None:
        text = "Just some text without metadata"
        decoded = decode_memory_text(text)

        assert decoded.content == text
        assert decoded.tags == []
        assert decoded.importance == MemoryImportance.NORMAL

    def test_decode_malformed_metadata(self) -> None:
        text = f"not-valid-json{MEMORY_METADATA_SEPARATOR}actual content"
        decoded = decode_memory_text(text)

        assert decoded.content == text
        assert decoded.tags == []
        assert decoded.importance == MemoryImportance.NORMAL

    def test_encode_empty_tags(self) -> None:
        encoded = encode_memory_text("test", [], MemoryImportance.LOW)
        decoded = decode_memory_text(encoded)

        assert decoded.content == "test"
        assert decoded.tags == []
        assert decoded.importance == MemoryImportance.LOW

    def test_all_importance_levels(self) -> None:
        for importance in MemoryImportance:
            encoded = encode_memory_text("test", [], importance)
            decoded = decode_memory_text(encoded)
            assert decoded.importance == importance

    def test_special_characters(self) -> None:
        content = 'Content with "quotes" and\nnewlines and {braces}'
        encoded = encode_memory_text(content, ["special"], MemoryImportance.NORMAL)
        decoded = decode_memory_text(encoded)
        assert decoded.content == content


# --- Mock Runtime (aligned with runtime DB API) ---


class MockRuntime:
    def __init__(
        self,
        memories: list[dict] | None = None,
        model_response: str | None = None,
    ) -> None:
        self._memories = memories or []
        self._model_response = model_response
        self.agent_id = "test-agent"
        self.create_memory = AsyncMock(return_value="mem-uuid")
        self.delete_memory = AsyncMock()

    async def get_memories(self, params: dict) -> list[dict]:
        return self._memories

    def get_setting(self, key: str) -> str | None:
        return None

    def get_service(self, name: str) -> object | None:
        return None

    async def use_model(self, model_type: str, params: dict) -> str | None:
        return self._model_response


# --- REMEMBER Action ---


class TestRememberAction:
    def test_metadata(self) -> None:
        assert remember_action.name == "REMEMBER"
        assert remember_action.description
        assert "remember" in remember_action.similes
        assert len(remember_action.examples) > 0

    @pytest.mark.asyncio
    async def test_validate_with_create_memory(self) -> None:
        runtime = MockRuntime()
        result = await remember_action.validate(runtime, {"content": {"text": "test"}})
        assert result is True

    @pytest.mark.asyncio
    async def test_validate_without_create_memory(self) -> None:
        runtime = MockRuntime()
        del runtime.create_memory
        result = await remember_action.validate(runtime, {"content": {"text": "test"}})
        assert result is False

    @pytest.mark.asyncio
    async def test_store_memory(self) -> None:
        runtime = MockRuntime(
            model_response='{"memory": "User likes dark mode", "tags": ["preference"], "importance": 2}',
        )
        message = {
            "agentId": "a1",
            "roomId": "r1",
            "userId": "u1",
            "content": {"text": "I like dark mode"},
        }

        result = await remember_action.handler(runtime, message)
        assert result["success"] is True
        assert "Remembered" in result["text"]
        runtime.create_memory.assert_called_once()


# --- RECALL Action ---


class TestRecallAction:
    def test_metadata(self) -> None:
        assert recall_action.name == "RECALL"
        assert recall_action.description
        assert "recall" in recall_action.similes

    @pytest.mark.asyncio
    async def test_no_memories(self) -> None:
        runtime = MockRuntime(memories=[])
        message = {"roomId": "r1", "content": {"text": "what do you remember?"}}

        result = await recall_action.handler(runtime, message)
        assert result["success"] is True
        assert "don't have any stored memories" in result["text"]

    @pytest.mark.asyncio
    async def test_find_matching_memory(self) -> None:
        encoded = encode_memory_text("Favorite color is blue", ["preference"], MemoryImportance.NORMAL)
        memories = [
            {
                "id": "m1",
                "agentId": "a1",
                "roomId": "r1",
                "content": {"text": encoded, "source": MEMORY_SOURCE},
                "createdAt": 1700000000000,
            }
        ]
        runtime = MockRuntime(memories=memories)
        message = {"roomId": "r1", "content": {"text": "color"}}

        result = await recall_action.handler(runtime, message)
        assert result["success"] is True
        assert "Found 1 memory" in result["text"]
        assert "Favorite color is blue" in result["text"]


# --- FORGET Action ---


class TestForgetAction:
    def test_metadata(self) -> None:
        assert forget_action.name == "FORGET"
        assert forget_action.description
        assert "forget" in forget_action.similes

    @pytest.mark.asyncio
    async def test_remove_by_id(self) -> None:
        runtime = MockRuntime()
        message = {"roomId": "r1", "content": {"text": "forget this"}}
        options = {"memoryId": "mem-123"}

        result = await forget_action.handler(runtime, message, None, options)
        assert result["success"] is True
        runtime.delete_memory.assert_called_once_with("mem-123")

    @pytest.mark.asyncio
    async def test_no_memories_to_remove(self) -> None:
        runtime = MockRuntime(memories=[])
        message = {"roomId": "r1", "content": {"text": "forget about colors"}}

        result = await forget_action.handler(runtime, message)
        assert result["success"] is True
        assert "No stored memories" in result["text"]


# --- Memory Context Provider ---


class TestMemoryContextProvider:
    def test_metadata(self) -> None:
        assert memory_context_provider.name == "MEMORY_CONTEXT"
        assert memory_context_provider.description

    @pytest.mark.asyncio
    async def test_empty_store(self) -> None:
        runtime = MockRuntime(memories=[])
        message = {"roomId": "r1", "content": {"text": ""}}

        result = await memory_context_provider.get(runtime, message, {})
        assert "No stored memories" in result.text

    @pytest.mark.asyncio
    async def test_with_memories(self) -> None:
        encoded = encode_memory_text("Test memory", ["test"], MemoryImportance.HIGH)
        memories = [
            {
                "id": "m1",
                "content": {"text": encoded, "source": MEMORY_SOURCE},
                "createdAt": 1000,
            }
        ]
        runtime = MockRuntime(memories=memories)
        message = {"roomId": "r1", "content": {"text": ""}}

        result = await memory_context_provider.get(runtime, message, {})
        assert "Stored Memories (1)" in result.text
        assert "Test memory" in result.text

    @pytest.mark.asyncio
    async def test_sorted_by_importance(self) -> None:
        low = encode_memory_text("Low item", [], MemoryImportance.LOW)
        high = encode_memory_text("High item", [], MemoryImportance.HIGH)
        memories = [
            {"id": "m1", "content": {"text": low, "source": MEMORY_SOURCE}, "createdAt": 1000},
            {"id": "m2", "content": {"text": high, "source": MEMORY_SOURCE}, "createdAt": 2000},
        ]
        runtime = MockRuntime(memories=memories)
        message = {"roomId": "r1", "content": {"text": ""}}

        result = await memory_context_provider.get(runtime, message, {})
        high_idx = result.text.index("High item")
        low_idx = result.text.index("Low item")
        assert high_idx < low_idx



"""
Auto-generated canonical action/provider/evaluator docs for plugin-memory.
DO NOT EDIT - Generated from prompts/specs/**.
"""

from __future__ import annotations

import json
from typing import TypedDict

class ActionDoc(TypedDict, total=False):
    name: str
    description: str
    similes: list[str]
    parameters: list[object]
    examples: list[list[object]]

class ProviderDoc(TypedDict, total=False):
    name: str
    description: str
    position: int
    dynamic: bool

class EvaluatorDoc(TypedDict, total=False):
    name: str
    description: str
    similes: list[str]
    alwaysRun: bool
    examples: list[object]

_CORE_ACTION_DOCS_JSON = """{
  "version": "1.0.0",
  "actions": []
}"""
_ALL_ACTION_DOCS_JSON = """{
  "version": "1.0.0",
  "actions": []
}"""
_CORE_PROVIDER_DOCS_JSON = """{
  "version": "1.0.0",
  "providers": []
}"""
_ALL_PROVIDER_DOCS_JSON = """{
  "version": "1.0.0",
  "providers": []
}"""
_CORE_EVALUATOR_DOCS_JSON = """{
  "version": "1.0.0",
  "evaluators": []
}"""
_ALL_EVALUATOR_DOCS_JSON = """{
  "version": "1.0.0",
  "evaluators": []
}"""

core_action_docs: dict[str, object] = json.loads(_CORE_ACTION_DOCS_JSON)
all_action_docs: dict[str, object] = json.loads(_ALL_ACTION_DOCS_JSON)
core_provider_docs: dict[str, object] = json.loads(_CORE_PROVIDER_DOCS_JSON)
all_provider_docs: dict[str, object] = json.loads(_ALL_PROVIDER_DOCS_JSON)
core_evaluator_docs: dict[str, object] = json.loads(_CORE_EVALUATOR_DOCS_JSON)
all_evaluator_docs: dict[str, object] = json.loads(_ALL_EVALUATOR_DOCS_JSON)

__all__ = [
    "ActionDoc",
    "ProviderDoc",
    "EvaluatorDoc",
    "core_action_docs",
    "all_action_docs",
    "core_provider_docs",
    "all_provider_docs",
    "core_evaluator_docs",
    "all_evaluator_docs",
]



"""RECALL action - retrieve stored memories based on a query or topic."""

from __future__ import annotations

import logging
from datetime import datetime, timezone

from elizaos_plugin_memory.actions.base import (
    ActionExample,
    ActionResult,
    HandlerCallback,
    Memory,
    RuntimeProtocol,
    State,
    create_action,
)
from elizaos_plugin_memory.types import (
    MEMORY_SOURCE,
    PLUGIN_MEMORY_TABLE,
    MemoryImportance,
    decode_memory_text,
)

logger = logging.getLogger(__name__)


async def validate(
    runtime: RuntimeProtocol,
    _message: Memory,
    _state: State | None = None,
) -> bool:
    try:
        return callable(getattr(runtime, "get_memories", None))
    except Exception:
        return False


async def handler(
    runtime: RuntimeProtocol,
    message: Memory,
    _state: State | None = None,
    options: dict[str, str | list[str] | int | None] | None = None,
    callback: HandlerCallback | None = None,
) -> ActionResult:
    try:
        content_data = message.get("content", {})
        content = content_data.get("text", "")
        if not content:
            error_message = "Please provide a query to recall memories."
            if callback:
                await callback({"text": error_message, "source": content_data.get("source")})
            return {"text": error_message, "success": False}

        query = str(options.get("query", content)) if options and options.get("query") else content
        filter_tags: list[str] = (
            list(options.get("tags", [])) if options and options.get("tags") else []
        )
        limit = int(options.get("limit", 10)) if options and options.get("limit") else 10
        min_imp = (
            int(options.get("minImportance", 1))
            if options and options.get("minImportance")
            else 1
        )

        memories = await runtime.get_memories(
            {
                "roomId": message.get("roomId", ""),
                "tableName": PLUGIN_MEMORY_TABLE,
                "count": 100,
            }
        )

        plugin_memories = [
            m for m in memories if m.get("content", {}).get("source") == MEMORY_SOURCE
        ]

        if not plugin_memories:
            no_mem_msg = "I don't have any stored memories yet."
            if callback:
                await callback({"text": no_mem_msg, "source": content_data.get("source")})
            return {"text": no_mem_msg, "success": True, "data": {"memories": [], "count": 0}}

        # Parse and filter
        parsed_list = []
        for m in plugin_memories:
            parsed = decode_memory_text(m.get("content", {}).get("text", ""))
            if int(parsed.importance) < min_imp:
                continue
            parsed_list.append(
                {
                    "id": m.get("id", ""),
                    "content": parsed.content,
                    "tags": parsed.tags,
                    "importance": parsed.importance,
                    "createdAt": m.get("createdAt", 0),
                }
            )

        if filter_tags:
            parsed_list = [
                m
                for m in parsed_list
                if any(tag in m["tags"] for tag in filter_tags)
            ]

        # Score and rank
        query_lower = query.lower()
        query_words = [w for w in query_lower.split() if len(w) >= 2]

        scored = []
        for m in parsed_list:
            content_lower = m["content"].lower()
            tags_str = " ".join(m["tags"]).lower()
            score = 0

            if query_lower in content_lower:
                score += 10

            for word in query_words:
                if word in content_lower:
                    score += 2
                if word in tags_str:
                    score += 3

            score += int(m["importance"])

            if score > 0:
                scored.append({**m, "score": score})

        scored.sort(key=lambda x: x["score"], reverse=True)
        scored = scored[:limit]

        if not scored:
            no_results_msg = "No memories found matching your query."
            if callback:
                await callback({"text": no_results_msg, "source": content_data.get("source")})
            return {
                "text": no_results_msg,
                "success": True,
                "data": {"memories": [], "count": 0},
            }

        lines = []
        for i, m in enumerate(scored):
            tag_str = f" [{', '.join(m['tags'])}]" if m["tags"] else ""
            created = m.get("createdAt", 0)
            date_str = (
                datetime.fromtimestamp(created / 1000, tz=timezone.utc).strftime("%Y-%m-%d")
                if created
                else "unknown"
            )
            lines.append(f"{i + 1}. {m['content']}{tag_str} ({date_str})")

        count = len(scored)
        suffix = "y" if count == 1 else "ies"
        result_text = f"Found {count} memor{suffix}:\n\n" + "\n".join(lines)

        if callback:
            await callback({"text": result_text, "source": content_data.get("source")})

        return {
            "text": result_text,
            "success": True,
            "data": {
                "memories": [
                    {
                        "id": m["id"],
                        "content": m["content"],
                        "tags": m["tags"],
                        "importance": int(m["importance"]),
                        "createdAt": m["createdAt"],
                    }
                    for m in scored
                ],
                "count": count,
            },
        }

    except Exception as error:
        logger.error("Failed to recall memories: %s", error)
        error_message = f"Failed to recall memories: {error}"
        if callback:
            await callback(
                {"text": error_message, "source": message.get("content", {}).get("source")}
            )
        return {"text": error_message, "success": False}


recall_action = create_action(
    name="RECALL",
    description="Retrieve stored memories based on a query, tags, or topic",
    similes=["recall", "remember-what", "search-memory", "find-memory", "what-do-you-remember"],
    examples=[
        [
            ActionExample(
                name="User",
                content={"text": "What do you remember about my preferences?"},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "Let me search my memories about your preferences.",
                    "actions": ["RECALL"],
                },
            ),
        ],
        [
            ActionExample(
                name="User",
                content={"text": "Recall everything about the project deadline."},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "I'll look up what I know about the project deadline.",
                    "actions": ["RECALL"],
                },
            ),
        ],
    ],
    validate=validate,
    handler=handler,
)



from elizaos_plugin_memory.actions.forget import forget_action
from elizaos_plugin_memory.actions.recall import recall_action
from elizaos_plugin_memory.actions.remember import remember_action

__all__ = [
    "remember_action",
    "recall_action",
    "forget_action",
]



"""Memory context provider - provides relevant long-term memories for conversation context."""

from __future__ import annotations

from elizaos_plugin_memory.providers.base import Provider, ProviderResult
from elizaos_plugin_memory.types import (
    IMPORTANCE_LABELS,
    MEMORY_SOURCE,
    PLUGIN_MEMORY_TABLE,
    decode_memory_text,
)


async def get_memory_context(runtime: object, message: object, _state: object) -> ProviderResult:
    try:
        get_memories = getattr(runtime, "get_memories", None)
        if not callable(get_memories):
            return ProviderResult(text="Runtime get_memories is not available")

        msg_dict = message if isinstance(message, dict) else {}
        room_id = msg_dict.get("roomId", "")

        memories = await get_memories(
            {"roomId": room_id, "tableName": PLUGIN_MEMORY_TABLE, "count": 50}
        )

        plugin_memories = [
            m for m in memories if m.get("content", {}).get("source") == MEMORY_SOURCE
        ]

        if not plugin_memories:
            return ProviderResult(text="No stored memories available")

        # Parse and sort by importance then recency
        parsed_list = []
        for m in plugin_memories:
            parsed = decode_memory_text(m.get("content", {}).get("text", ""))
            parsed_list.append(
                {
                    "id": m.get("id", ""),
                    "content": parsed.content,
                    "tags": parsed.tags,
                    "importance": parsed.importance,
                    "createdAt": m.get("createdAt", 0),
                }
            )

        parsed_list.sort(
            key=lambda x: (int(x["importance"]), x["createdAt"]), reverse=True
        )
        parsed_list = parsed_list[:20]

        lines = []
        for m in parsed_list:
            tag_str = f" [{', '.join(m['tags'])}]" if m["tags"] else ""
            level = IMPORTANCE_LABELS.get(int(m["importance"]), "normal")
            lines.append(f"- ({level}) {m['content']}{tag_str}")

        count = len(parsed_list)
        text = f"Stored Memories ({count}):\n" + "\n".join(lines)

        return ProviderResult(
            text=text,
            data={
                "memories": [
                    {
                        "id": m["id"],
                        "content": m["content"],
                        "tags": m["tags"],
                        "importance": int(m["importance"]),
                    }
                    for m in parsed_list
                ],
                "count": count,
            },
        )
    except Exception:
        return ProviderResult(text="Error retrieving stored memories")


memory_context_provider = Provider(
    name="MEMORY_CONTEXT",
    description="Provides relevant long-term memories from conversation context",
    get=get_memory_context,
)



"""REMEMBER action - store a piece of information as a long-term memory."""

from __future__ import annotations

import json
import logging
import re

from elizaos_plugin_memory.actions.base import (
    ActionExample,
    ActionResult,
    HandlerCallback,
    Memory,
    RuntimeProtocol,
    State,
    create_action,
)
from elizaos_plugin_memory.types import (
    MEMORY_SOURCE,
    PLUGIN_MEMORY_TABLE,
    MemoryImportance,
    encode_memory_text,
)

logger = logging.getLogger(__name__)


async def validate(
    runtime: RuntimeProtocol,
    _message: Memory,
    _state: State | None = None,
) -> bool:
    try:
        return callable(getattr(runtime, "create_memory", None))
    except Exception:
        return False


async def handler(
    runtime: RuntimeProtocol,
    message: Memory,
    _state: State | None = None,
    options: dict[str, str | list[str] | int | None] | None = None,
    callback: HandlerCallback | None = None,
) -> ActionResult:
    try:
        content_data = message.get("content", {})
        content = content_data.get("text", "")
        if not content:
            error_message = "Please provide content to remember."
            if callback:
                await callback({"text": error_message, "source": content_data.get("source")})
            return {"text": error_message, "success": False}

        memory_text = str(options.get("content", content)) if options else content
        tags: list[str] = list(options.get("tags", [])) if options and options.get("tags") else []
        importance_val = int(options.get("importance", 2)) if options and options.get("importance") else 2
        importance = (
            MemoryImportance(importance_val)
            if 1 <= importance_val <= 4
            else MemoryImportance.NORMAL
        )

        # Use LLM to extract structured memory if no explicit content given
        if not options or not options.get("content"):
            prompt = (
                "Extract the key information to remember from this message.\n"
                "Return ONLY a JSON object (no markdown, no code blocks):\n"
                '{"memory": "The concise fact to store", '
                '"tags": ["relevant", "tags"], "importance": 2}\n\n'
                "Importance: 1=low, 2=normal, 3=high, 4=critical\n\n"
                f'User message: "{content}"'
            )

            response = await runtime.use_model("TEXT_LARGE", {"prompt": prompt})
            if response:
                try:
                    cleaned = re.sub(r"^```(?:json)?\n?", "", response)
                    cleaned = re.sub(r"\n?```$", "", cleaned).strip()
                    parsed = json.loads(cleaned)
                    memory_text = parsed.get("memory", content)
                    raw_tags = parsed.get("tags", [])
                    tags = [str(t) for t in raw_tags] if isinstance(raw_tags, list) else []
                    imp = parsed.get("importance", 2)
                    importance = (
                        MemoryImportance(int(imp))
                        if isinstance(imp, (int, float)) and 1 <= int(imp) <= 4
                        else MemoryImportance.NORMAL
                    )
                except (json.JSONDecodeError, ValueError) as parse_err:
                    logger.warning("Failed to parse memory extraction: %s", parse_err)
                    memory_text = content

        encoded = encode_memory_text(memory_text, tags, importance)

        entity_id = message.get("entityId") or message.get("userId", "")
        memory_entry: Memory = {
            "agentId": runtime.agent_id,
            "roomId": message.get("roomId", ""),
            "userId": entity_id,
            "content": {"text": encoded, "source": MEMORY_SOURCE},
            "createdAt": 0,
        }

        await runtime.create_memory(memory_entry, PLUGIN_MEMORY_TABLE, True)

        tag_suffix = f" [tags: {', '.join(tags)}]" if tags else ""
        success_message = f'Remembered: "{memory_text}"{tag_suffix}'

        if callback:
            await callback({"text": success_message, "source": content_data.get("source")})

        return {
            "text": success_message,
            "success": True,
            "data": {"content": memory_text, "tags": tags, "importance": int(importance)},
        }

    except Exception as error:
        logger.error("Failed to store memory: %s", error)
        error_message = f"Failed to store memory: {error}"
        if callback:
            await callback(
                {"text": error_message, "source": message.get("content", {}).get("source")}
            )
        return {"text": error_message, "success": False}


remember_action = create_action(
    name="REMEMBER",
    description="Store a piece of information as a long-term memory for later recall",
    similes=["remember", "memorize", "store-memory", "save-memory", "note-down"],
    examples=[
        [
            ActionExample(
                name="User",
                content={"text": "Remember that my favorite color is blue."},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "I'll remember that your favorite color is blue.",
                    "actions": ["REMEMBER"],
                },
            ),
        ],
        [
            ActionExample(
                name="User",
                content={"text": "Memorize this: the project deadline is March 15th."},
            ),
            ActionExample(
                name="Assistant",
                content={
                    "text": "Got it, I've stored that the project deadline is March 15th.",
                    "actions": ["REMEMBER"],
                },
            ),
        ],
    ],
    validate=validate,
    handler=handler,
)
